import importlib
import sys
import inspect

from unittest.mock import MagicMock

_stdout_pointer = 0


def load_module(pkg, lang='en'):
    """
    Used to load a module without::
        def main()
            pass
        if __name__ == "__main__":
            main()
    When loaded, runs the code immediately.
    """
    module_not_found = 'File {0} does not exist!'.format(pkg)
    other_exception = 'Running exercise {0} failed. Please make sure that you can run your code.'.format(pkg)
    exit_called = 'Make sure your program does not exit with an exit() command.'

    if lang == 'fi':
        module_not_found = 'Tiedostoa {0} ei löytynyt.'.format(pkg)
        other_exception = 'Tehtävän {0} suorittaminen epäonnistui. '.format(pkg) \
            + 'Varmista, että saat ohjelman suoritettua loppuun.'
        exit_called = 'Varmista, että koodisi ei kutsu exit() komentoa.'

    try:
        return importlib.import_module(pkg)
    except ModuleNotFoundError:
        return AssertionError(module_not_found)
    except Exception:
        return AssertionError(other_exception)
    except SystemExit:
        return AssertionError(exit_called)


def reload_module(module):
    """Runs the module code again, used when no main() defined"""
    if isinstance(module, AssertionError):
        raise module
    clear_stdout()
    importlib.reload(module)


def load(pkg, method, lang='en', err=None):
    """
    Loads a method from a module, doesn't run the code, needs to be called in tests.
    Exercise Example::
        import numpy as np
        def main():
            [print(line) for line in range(4)]
    Test Example::
        module_name="src.filename"
        main = load(module_name, "main")
        def test_lines(self):
            main()
            result = get_stdout().split('\\n')
            self.assertEqual(len(result), 4, msg="The output should contain exactly four lines!")
    """
    module_not_found = 'Function {1} was not found in file {0}.'.format(pkg, method)
    if lang == 'fi':
        module_not_found = 'Tiedostosta {0} ei löytynyt funktiota {1}.'.format(pkg, method)

    if not err:
        err = module_not_found

    def fail(*args, **kwargs):
        if args:
            raise AssertionError(args[0])
        raise AssertionError(err)

    try:
        return getattr(importlib.import_module(pkg), method)
    except ModuleNotFoundError as mnf:
        return fail(mnf)
    except Exception as e:
        return fail


def get_stdout():
    global _stdout_pointer
    return sys.stdout.getvalue()[_stdout_pointer:].strip()


def get_stderr():
    return sys.stderr.getvalue().strip()


def any_contains(needle, haystacks):
    any(map(lambda haystack: needle in haystack, haystacks))


def clear_stdout():
    global _stdout_pointer
    _stdout_pointer = len(sys.stdout.getvalue())


def check_source(module):
    """
    Check that module doesn't have any globals.
    Example::
        def test_no_global(self):
            result, line = check_source(self.module)
            self.assertTrue(result, "Make sure no code is outside functions.\\nRow: " + line)
    """
    try:
        source = inspect.getsource(module)
    except Exception:
        raise Exception('Varmista, että koodin suoritus onnistuu')
    allowed = [
        "import ",
        "from ",
        "def ",
        "class ",
        " ",
        "\t",
        "#",
        "if __name__",
        "@",
    ]

    for line in source.split("\n"):
        if line.strip() == "":
            continue
        for prefix in allowed:
            if line.startswith(prefix):
                break
        else:
            return (False, line)
    return (True, "")


def remove_extra_whitespace(mj):
    mj = mj.strip()
    while "  " in mj:
        mj = mj.replace("  ", " ")
    return mj


def sanitize(mj):
    """
    Sanitize string, remove all unnecessary whitespaces.
    """
    return '\n'.join([remove_extra_whitespace(m) for m in mj.split('\n')])


def assert_ignore_ws(self, was, expected, errmsg='', lang='fi'):
    """
    Assert Ignore all whitespace in output.
    Example::
        assert_ignore_ws(self, output[0], 'Ukko Nooa', "First line doesn't match. ")
    """
    xmj1 = ''.join([x for x in remove_extra_whitespace(was).split(' ') if len(x) > 0])
    xmj2 = ''.join([x for x in remove_extra_whitespace(expected).split(' ') if len(x) > 0])
    if lang == 'fi':
        err = '{0}\nTulostit:\n{1}\nOdotettiin:\n{2}'.format(errmsg, was, expected)
    else:
        err = '{0}\nYou printed:\n{1}\nExpected:\n{2}'.format(errmsg, was, expected)
    self.assertTrue(xmj1 == xmj2, err)


def spy_decorator(method_to_decorate, name):
    """
    This solution to wrap a patched method comes originally from
    https://stackoverflow.com/questions/25608107/python-mock-patching-a-method-without-obstructing-implementation
    """
    mock = MagicMock(name="%s method" % name)

    def wrapper(self, *args, **kwargs):
        mock(*args, **kwargs)
        return method_to_decorate(self, *args, **kwargs)
    wrapper.mock = mock
    return wrapper


class patch_helper(object):
    """
    patch_helper code copied from Data Analysis with Python.
    Example::
        from tmc.utils import load, get_stdout, patch_helper
        module_name='src.file_listing'
        ph = patch_helper(module_name)
    In tests file, if you want to patch "src.file_listing.re.compile" use following:
    Example::
        def test_content(self):
            patch(ph('re.compile'), side_effect=re.compile) as c:
                ...
    """

    def __init__(self, module_name):
        import importlib
        self.m = module_name

    def __call__(self, d):
        # import importlib
        parts = d.split(".")
        # If e.g. d == package.subpackage.subpackage2.attribute,
        # and our module is called mystery_data.
        try:
            getattr(importlib.import_module(self.m), parts[-1])   # attribute
            p = ".".join([self.m, parts[-1]])
            # p='src.mystery_data.attribute'
        except ModuleNotFoundError:
            raise
        except AttributeError:
            if len(parts) == 1:
                raise
            try:
                getattr(importlib.import_module(self.m), parts[-2])  # subpackage2.attribute
                p = ".".join([self.m] + parts[-2:])
                # p='src.mystery_data.subpackage2.attribute'
            except AttributeError:
                if len(parts) == 2:
                    raise
                try:
                    getattr(importlib.import_module(self.m), parts[-3])  # subpackage.subpackage2.attribute
                    p = ".".join([self.m] + parts[-3:])
                    # p='src.mystery_date.subpackage.subpackage2.attribute'
                except AttributeError:
                    if len(parts) == 3:
                        raise
                    # package.subpackage.subpackage2.attribute
                    getattr(importlib.import_module(self.m), parts[-4])
                    p = ".".join([self.m] + parts[-4:])
                    # p='src.mystery_date.package.subpackage.subpackage2.attribute'
        return p



from pathlib import Path

import pytest

import testing.util
from tmc_course import tmc_course


@pytest.fixture
def test_resource_path() -> Path:
    return testing.util.test_resource_dir()


@pytest.fixture
def tmp_course(tmp_path) -> Path:
    course_path = tmp_path / "NewCourse"
    tmc_course.init_course(course_path)
    return course_path


@pytest.fixture
def tmp_part(tmp_course) -> Path:
    tmc_course.init_part(tmp_course, "part01")
    return tmp_course / "part01"



from unittest import TestProgram
from .runner import TMCTestRunner
import sys
from tmc import result

django_defined = False
try:
    with open('.tmcproject.yml') as f:
        for line in f:
            try:
                (key, value) = line.split(":")
                if key.strip().lower() == "django":
                    django_defined = str(value.strip().lower())
            except ValueError:
                pass
except FileNotFoundError:
    pass

if django_defined:
    import os
    import django
    import django.conf
    from django.test.utils import get_runner
    from django.conf import settings
    os.environ['DJANGO_SETTINGS_MODULE'] = django_defined + '.config.settings'
    django.setup()

if sys.argv.__len__() > 1 and sys.argv[1] == 'available_points':
    TMCTestRunner().available_points()
    sys.exit()

if '--wait-for-secret' in sys.argv:
    sys.argv.remove("--wait-for-secret")
    result.module_secret = input("waiting for a secret ")

if django_defined:
    settings.TEST_RUNNER = 'tmc.django.TMCDiscoverRunner'
    TestRunner = get_runner(settings)
    test_runner = TestRunner()
    failures = test_runner.run_tests(["test"])
    sys.exit(bool(failures))

main = TestProgram
main(testRunner=TMCTestRunner, module=None, failfast=False, buffer=True)



from inspect import isclass, isfunction
from collections import defaultdict

point_register = {'suite': defaultdict(list), 'test': defaultdict(list)}


def qualifier(test):
    return "%s.%s" % (test.__module__, test.__qualname__)


def save_points(o, points, dst):
    q = qualifier(o)
    dst[q] += filter(lambda point: point not in dst[q], points)


def points(*points):

    def points_wrapper(o):
        if isclass(o):
            save_points(o, points, point_register['suite'])
        elif isfunction(o):
            save_points(o, points, point_register['test'])
        else:
            raise Exception("Expected decorator object '%s' type to be Class or Function but was %s." % (o, type(o)))
        return o

    if not points:
        raise Exception("You need to define at least one point in the points decorator declaration")
    for point in points:
        if type(point) is not str:
            msg = "Points decorator argument '%s' needs to be a string, but was %s." % (point, type(point).__name__)
            raise Exception(msg)
    return points_wrapper


def _parse_points(test):
    name = _name_test(test)
    testPoints = point_register['test']
    points = testPoints[name]
    key = name[:name.rfind('.')]
    suitePoints = point_register['suite'][key]
    points += suitePoints
    return points


def _name_test(test):
    module = test.__module__
    classname = test.__class__.__name__
    testName = test._testMethodName
    return module + '.' + classname + '.' + testName



import filecmp
from pathlib import Path
from typing import Optional


def test_resource_dir() -> Path:
    return Path(__file__).parent / "resources"


def normalized_filecmp(expected_file: Path, actual_file: Path) -> bool:
    """Compares two files ignoring the type of the newline, because TMC does it's
    own magic down the line anyways.

    It should be enough to open() both files, as python apparently just converts
    to universal-mode newlines at that point. For binary files, we can fall back
    to filecmp.cmp.
    """
    with expected_file.open("r") as expected_fh, actual_file.open("r") as actual_fh:
        try:
            expected_lines = expected_fh.readlines()
            actual_lines = actual_fh.readlines()
        except UnicodeDecodeError:
            # Binary files, just filecmp.cmp because there are no newlines at this point
            return filecmp.cmp(expected_file, actual_file)
    if len(expected_lines) != len(actual_lines):
        return False
    for expected_line, actual_line in zip(expected_lines, actual_lines):
        if expected_line != actual_line:
            return False
    return True


def assert_dir_equals(
    expected_path: Path, actual_path: Path, ignore: Optional[list[str]] = None
) -> None:
    # filecmp.dircmp is silly and only checks the os.stat() signature, which means it
    # ignores contents but enforces modification time. This means both a risk of
    # false positives and false negatives, which is impressive on it's own.
    expected_files = list(expected_path.iterdir())
    actual_files = list(actual_path.iterdir())

    if ignore:
        expected_files = [f for f in expected_files if f.name not in ignore]
        actual_files = [f for f in actual_files if f.name not in ignore]

    expected_filenames = set(f.name for f in expected_files)
    actual_filenames = set(f.name for f in actual_files)
    assert expected_filenames == actual_filenames, (
        "File name sets are not equivalent; ",
        f"{expected_filenames=}, {actual_filenames=}; ",
        f"diff: {expected_filenames.symmetric_difference(actual_filenames)}",
    )

    # Doing what filecmp.dircmp *should* be doing
    expected_files = sorted(expected_files)
    actual_files = sorted(actual_files)
    for expected_file, actual_file in zip(expected_files, actual_files):
        assert (
            expected_file.name == actual_file.name
        ), f"File names differ; {expected_file.name=}, {actual_file.name=}"
        if expected_file.is_dir():
            assert_dir_equals(expected_file, actual_file)
        else:
            assert normalized_filecmp(
                expected_file, actual_file
            ), f"File contents differ; {expected_file=}, {actual_file=}"



import unittest

from tmc import points
from tmc.utils import check_source, load, load_module

exercise = "src.solution"


@points("valid_assignment_en")
class SolutionTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.module = load_module(exercise, "en")

    def test_0_main_ok(self):
        ok, line = check_source(self.module)
        self.assertTrue(
            ok,
            (
                "Extra code for testing should reside within the \n"
                'if __name__ == "__main__":\n'
                "block. The following line must be moved:\n"
                "{line}"
            ),
        )

    def test_1(self):
        function = load(exercise, "function", "en")
        self.assertEquals(
            1, function(), "The function 'function' should return the value 1"
        )


if __name__ == "__main__":
    unittest.main()



import importlib
from bdb import Bdb
import sys


class RecursionDetected(Exception):
    pass


class RecursionDetector(Bdb):
    def do_clear(self, arg):
        pass

    def __init__(self, *args):
        Bdb.__init__(self, *args)
        self.stack = set()

    def user_call(self, frame, argument_list):
        code = frame.f_code
        if code in self.stack:
            raise RecursionDetected
        self.stack.add(code)

    def user_return(self, frame, return_value):
        self.stack.remove(frame.f_code)


def test_recursion(func: callable, *args):
    detector = RecursionDetector()
    detector.set_trace()
    try:
        func(*args)
    except RecursionDetected:
        return True
    else:
        return False
    finally:
        sys.settrace(None)


class Reflect:
    def __init__(self, modulename: str = "", classname: str = ""):
        self.__classname = classname
        self.__modulename = modulename
        self.__cls = None

    def load_class(self):
        try:
            self.__cls = getattr(importlib.import_module(self.__modulename), self.__classname)
            return self.__cls
        except Exception as e:
            return None

    def load_object(self, *params):
        try:
            if not self.__cls:
                self.load_class()
            self.__obj = self.cls(*params)
            return self.__obj
        except Exception as e:
            print(e)
            return None

    def set_object(self, obj):
        self.__obj = obj

    @property
    def cls(self):
        return self.__cls

    @property
    def object(self):
        return self.__obj

    def list_attributes(self, filter_builtin=False):
        if filter_builtin:
            return [x for x in dir(self.__obj) if not x.startswith("__")]
        return dir(self.__obj)

    def has_attribute(self, attribute: str):
        if attribute in dir(self.__obj):
            return True
        if ("_" + self.__classname + attribute) in dir(self.__obj):
            return True
        return False

    def get_attribute(self, attribute: str):
        if attribute in dir(self.__obj):
            return getattr(self.__obj, attribute)
        elif ("_" + self.__classname + attribute) in dir(self.__obj):
            return getattr(self.__obj, "_" + self.__classname + attribute)
        return None

    def list_public_members(self):
        return [x for x in dir(self.__obj) if not x.startswith("_")]



import unittest

from tmc import points
from tmc.utils import check_source, load, load_module

exercise = "src.ratkaisu"


@points("valid_assignment_fi")
class RatkaisuTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.module = load_module(exercise, "fi")

    def test_0_paaohjelma_kunnossa(self):
        ok, line = check_source(self.module)
        self.assertTrue(
            ok,
            (
                "Funktioita testaava koodi tulee sijoittaa lohkon\n"
                'if __name__ == "__main__":\n'
                "sisälle. Seuraava rivi tulee siirtää:\n"
                "{line}"
            ),
        )

    def test_1(self):
        funktio = load(exercise, "funktio", "fi")
        self.assertEquals(1, funktio(), "Funktion 'funktio' tulisi palauttaa arvo 1")


if __name__ == "__main__":
    unittest.main()



from tmc.hmac_writer import write_hmac
from unittest.runner import TextTestResult
from .points import _parse_points, _name_test
from copy import deepcopy
import atexit
import gc
import json
import traceback

module_secret = None
# Copy of the results for the Python editor
results = []

class TMCResult(TextTestResult):

    def __init__(self, stream, descriptions, verbosity):
        self.__results = []
        global module_secret
        secret = module_secret
        del module_secret
        that = self

        def write_output():
            nonlocal that
            nonlocal secret
            output = json.dumps(that.__results)
            if secret is not None:
                write_hmac(secret, output)
            with open(".tmc_test_results.json", "w") as text_file:
                text_file.write(output)
        atexit.register(write_output)
        super(TMCResult, self).__init__(stream, descriptions, verbosity)

    def startTest(self, test):
        # It is easy to use a lot of memory in a unit test and if we don't force garbage collection here, we can go over the memory limit in the sandbox
        gc.collect()
        super(TMCResult, self).startTest(test)

    def stopTest(self, test):
        super(TMCResult, self).stopTest(test)
        # It is easy to use a lot of memory in a unit test and if we don't force garbage collection here, we can go over the memory limit in the sandbox
        gc.collect()

    def addSuccess(self, test):
        super(TMCResult, self).addSuccess(test)
        self.addResult(test, 'passed')

    def addFailure(self, test, err):
        super(TMCResult, self).addFailure(test, err)
        self.addResult(test, 'failed', err)

    def addError(self, test, err):
        super(TMCResult, self).addError(test, err)
        self.addResult(test, 'errored', err)

    def addResult(self, test, status, err=None):
        global results
        points = _parse_points(test)
        message = ""
        backtrace = []
        if err is not None:
            message = str(err[1])
            backtrace = traceback.format_tb(err[2])

        details = {
            'name': _name_test(test),
            'status': status,
            'message': message,
            'passed': status == 'passed',
            'points': points,
            'backtrace': backtrace
        }
        self.__results.append(details)
        results = deepcopy(self.__results)



import argparse
import contextlib
import importlib.metadata
import importlib.resources
import logging
import os
import shutil
import subprocess
import zipfile
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Generator, Literal, Optional

import requests
import treelib  # type: ignore
from tqdm import tqdm

TMC_PYTHON_TESTER_ZIP_URL = (
    "https://github.com/testmycode/tmc-python-tester/archive/refs/heads/master.zip"
)


class ActionCancelledException(BaseException):
    pass


class SkeletonFile(Enum):
    COURSE_GITIGNORE = auto()
    COURSE_TMCPROJECT_YML = auto()
    ASSIGNMENT_TMCPROJECT_YML = auto()
    ASSIGNMENT_SOLUTION_EN = auto()
    ASSIGNMENT_SOLUTION_FI = auto()
    ASSIGNMENT_TEST_EN = auto()
    ASSIGNMENT_TEST_FI = auto()


def check_from_user(
    message: str, quit_default: bool = True, continue_on_y: bool = True
) -> None:
    quit_char = "n" if continue_on_y else "y"
    continue_char = "y" if continue_on_y else "n"
    suffix = "["
    suffix += continue_char if quit_default else continue_char.upper()
    suffix += quit_char.upper() if quit_default else quit_char
    suffix += "]"

    print(message + " " + suffix)

    response = input("> ").casefold()
    if response == quit_char:
        raise ActionCancelledException
    elif response != continue_char and quit_default:
        raise ActionCancelledException


def add_skeleton_file(skeleton_file: SkeletonFile, path: Path) -> None:
    logging.debug(f"Copying resource {skeleton_file.name} to {path}")
    template_file_name = skeleton_file.name.casefold() + ".template"
    resource_path = importlib.resources.files("tmc_course.resources").joinpath(
        template_file_name
    )
    logging.debug(f"Template file path is {resource_path}")
    shutil.copy(Path(str(resource_path)), path)


def init_course(course_path: Path) -> None:
    logging.info(f"Initializing a new course in {course_path}")
    if course_path.exists():
        check_from_user(
            f"Directory {course_path} already exists. Continue and overwrite?"
        )

    if not course_path.name.replace("_", "").isalnum():
        raise ValueError("Course name must be alphanumeric (underscores allowed)")

    logging.debug(f"Creating directory {course_path}")
    course_path.mkdir(exist_ok=True, parents=True)

    add_skeleton_file(SkeletonFile.COURSE_GITIGNORE, course_path / ".gitignore")
    add_skeleton_file(
        SkeletonFile.COURSE_TMCPROJECT_YML, course_path / ".tmcproject.yml"
    )


def is_valid_course(course_path: Path) -> bool:
    if not course_path.exists():
        logging.debug("Course root directory does not exist")
        return False
    if not course_path.is_dir():
        logging.debug("Course root is not a directory")
        return False
    if not any(
        filepath.name == ".tmcproject.yml" for filepath in course_path.iterdir()
    ):
        logging.debug(
            "Course root does not appear to be a TMC course (missing .tmcproject.yml)"
        )
        return False
    logging.debug(f"{course_path} is a valid TMC course")
    return True


def init_part(course_path: Path, part_name: str) -> None:
    if not is_valid_course(course_path):
        raise ValueError(f"{course_path} is not a TMC course")

    if not part_name.replace("_", "").isalnum():
        raise ValueError("Part name must be alphanumeric (underscores allowed)")

    part_path = course_path / part_name
    logging.info(f"Initializing a new part in {part_path}")

    if part_path.exists():
        check_from_user(
            f"Directory {part_path} already exists. Continue and overwrite?"
        )
    logging.debug(f"Creating directory {part_path}")
    part_path.mkdir(exist_ok=True)


def is_valid_part(part: Path) -> bool:
    if not part.exists():
        logging.debug("Part {part_name} does not exist")
        return False
    if not part.is_dir():
        logging.debug(f"{part} is not a directory")
        return False
    if not is_valid_course(part.parent):
        logging.debug(f"Parent {part.parent} is not valid course")
        return False
    logging.debug(f"{part} is a valid course part")
    return True


def create_src_skeleton(assignment_path: Path, language: Literal["en", "fi"]) -> None:
    if language not in ("fi", "en"):
        raise ValueError("Language must be 'fi' or 'en'")

    logging.debug(f'Creating " {assignment_path / "src"}')
    (assignment_path / "src").mkdir(exist_ok=True)

    logging.debug('Creating {assignment_path / "src" / "__init__.py"}"')
    (assignment_path / "src" / "__init__.py").touch(exist_ok=True)

    if language == "en":
        add_skeleton_file(
            SkeletonFile.ASSIGNMENT_SOLUTION_EN,
            assignment_path / "src" / "solution.py",
        )
    else:
        add_skeleton_file(
            SkeletonFile.ASSIGNMENT_SOLUTION_FI,
            assignment_path / "src" / "ratkaisu.py",
        )


def create_test_skeleton(
    assignment_path: Path, assignment_name: str, language: Literal["en", "fi"]
) -> None:
    logging.debug(f'Creating {assignment_path / "test"}')
    (assignment_path / "test").mkdir(exist_ok=True)

    logging.debug(f'Creating {assignment_path / "test" / "__init__.py"}')
    (assignment_path / "test" / "__init__.py").touch(exist_ok=True)

    if language == "en":
        test_file_path = assignment_path / "test" / "test_solution.py"
        add_skeleton_file(SkeletonFile.ASSIGNMENT_TEST_EN, test_file_path)
    elif language == "fi":
        test_file_path = assignment_path / "test" / "test_ratkaisu.py"
        add_skeleton_file(SkeletonFile.ASSIGNMENT_TEST_FI, test_file_path)
    else:
        raise ValueError("Language must be 'fi' or 'en'")

    logging.debug("Inserting assignment name into test file")
    with test_file_path.open("r") as fh:
        file_contents = fh.read()
    file_contents = file_contents.replace("POINTNAME", assignment_name)
    with test_file_path.open("w") as fh:
        fh.write(file_contents)
    logging.debug("Test skeleton complete")


def download_tmc_python_tester(course_path: Path, update: bool) -> None:
    tester_zip_path = course_path / "tmc-python-tester.zip"
    logging.debug(
        f'Looking for TMC-python-tester zip at {course_path / "tmc-python-tester.zip"}'
    )

    if update or not tester_zip_path.exists():
        logging.info("Downloading TMC-python-tester")
        logging.debug(f"URL: {TMC_PYTHON_TESTER_ZIP_URL}")
        response = requests.get(TMC_PYTHON_TESTER_ZIP_URL, stream=True)
        with open(tester_zip_path, "wb") as fh:
            for chunk in response.iter_content(chunk_size=128):
                fh.write(chunk)


def create_tmc_dir(assignment_path: Path) -> None:
    logging.debug(f'Creating {assignment_path / "tmc"}')
    (assignment_path / "tmc").mkdir(exist_ok=True)

    course_path = assignment_path.parent.parent
    download_tmc_python_tester(course_path, update=False)

    with zipfile.ZipFile(course_path / "tmc-python-tester.zip") as tester_zip:
        for file_info in tester_zip.infolist():
            logging.debug(f"Looking at file {file_info.filename=}")
            if file_info.filename.startswith("tmc-python-tester-master/tmc/"):
                # Need to remove prefix, s.t. we don't retain the parent folders
                file_info.filename = file_info.filename.replace(
                    "tmc-python-tester-master/tmc", ""
                )
                logging.debug(
                    f'Extracting {file_info.filename} to {assignment_path / "tmc"}'
                )
                tester_zip.extract(file_info, assignment_path / "tmc")


def init_assignment(
    course_path: Path,
    part_name: str,
    assignment_name: str,
    language: Literal["fi", "en"],
) -> None:
    if not is_valid_course(course_path):
        raise ValueError(f"{course_path} is not a valid TMC course")
    if not is_valid_part(course_path / part_name):
        raise ValueError(f"{course_path / part_name} is not a valid course part")

    if not assignment_name.replace("_", "").isalnum():
        raise ValueError("Assignment name must be alphanumeric (underscores allowed)")

    assignment_path = course_path / part_name / assignment_name
    logging.info(f"Initializing a new assignment in {assignment_name}")

    if assignment_path.exists():
        check_from_user(
            f"Directory {assignment_path} already exists. Continue and overwrite?"
        )
    logging.debug(f"Creating {assignment_path}")
    assignment_path.mkdir(exist_ok=True)

    add_skeleton_file(
        SkeletonFile.ASSIGNMENT_TMCPROJECT_YML, assignment_path / ".tmcproject.yml"
    )
    create_src_skeleton(assignment_path, language)
    create_test_skeleton(assignment_path, assignment_name, language)
    create_tmc_dir(assignment_path)


def is_valid_assignment(assignment_path: Path) -> bool:
    if not assignment_path.exists():
        logging.debug(f"Assignment {assignment_path} does not exist")
        return False
    if not assignment_path.is_dir():
        logging.debug(f"Assignment {assignment_path} is not a directory")
        return False
    for tgt in (".tmcproject.yml", "test", "src", "tmc"):
        if not any(filepath.name == tgt for filepath in assignment_path.iterdir()):
            logging.debug(f"{assignment_path} is not a TMC assignment (missing {tgt})")
            return False
    logging.debug(f"{assignment_path} is a valid TMC assignment")
    return True


def update_course(course_path: Path) -> None:
    logging.info(f"Updating TMC-python-tester for course {course_path}")
    download_tmc_python_tester(course_path, update=True)
    is_valid_course(course_path)
    for maybe_part in course_path.iterdir():
        logging.debug(f"Checking whether {maybe_part} is a course part")
        if not is_valid_part(maybe_part):
            logging.debug("Not a part, skipping")
            continue
        for maybe_assignment in maybe_part.iterdir():
            logging.debug(f"Checking whether {maybe_assignment} is an assignment")
            if not is_valid_assignment(maybe_assignment):
                logging.debug("Not an assignment, skipping")
                continue
            logging.info(f"Updating assignment at {maybe_assignment}")
            create_tmc_dir(maybe_assignment)


@dataclass
class TestTask:
    path: Path

    @property
    def course_path(self) -> Path:
        return self.path.parent.parent

    @property
    def part_path(self) -> Path:
        return self.path.parent


@dataclass
class TestResult:
    task: TestTask
    success: bool
    stdout: str
    stderr: str


def collect_tasks(paths: list[Path]) -> Generator[TestTask, None, None]:
    for path in paths:
        if is_valid_assignment(path):
            logging.debug(f"{path} is assignment")
            yield TestTask(path)
        elif is_valid_part(path):
            logging.debug(f"{path} appears to be a part")
            yield from collect_tasks(
                list(child for child in path.iterdir() if child.is_dir())
            )
        elif is_valid_course(path):
            logging.debug(f"{path} appears to be a course")
            yield from collect_tasks(
                list(child for child in path.iterdir() if child.is_dir())
            )
        else:
            logging.debug(f"{path} is neither an assignment, a part, or a course")


def is_last_child_of_parent(nodeid: object, tree: treelib.Tree) -> bool:
    if not tree.parent(nodeid):
        return True

    # variable needed for mypy; treelib is untyped, so otherwise the result
    # of this __eq__ is Any
    result: bool = (
        tree.get_node(nodeid) == tree.children(tree.parent(nodeid).identifier)[-1]
    )
    return result


def print_test_output(results: list[TestResult]) -> None:
    tree = treelib.Tree()
    tree.create_node("Test Results", "root")

    for course_path in set(result.task.course_path for result in results):
        tree.create_node(course_path.name, course_path, parent="root")

    for part_path in set(result.task.part_path for result in results):
        tree.create_node(part_path.name, part_path, parent=part_path.parent)

    for idx, result in enumerate(results):
        affix = (
            "\x1b[32;1mSUCCESS\x1b[0m" if result.success else "\x1b[31;1mFAIL\x1b[0m"
        )
        tree.create_node(
            f"{result.task.path.name} - {affix}",
            result.task.path,
            parent=result.task.part_path,
        )

    tree.show()


def test(paths: list[Path], detailed: bool = False) -> tuple[bool, list[TestResult]]:
    logging.debug("Collecting assignments")
    tasks: list[TestTask] = list(collect_tasks(paths))

    results: list[TestResult] = []
    logging.debug("Running tests")
    for task in tqdm(
        tasks, unit=" assg", disable=not logging.getLogger().isEnabledFor(logging.INFO)
    ):
        results.append(run_test_task(task))

    for result in results:
        if detailed or not result.success:
            logging.info("")
            logging.info(f"\nTEST RESULTS FOR {result.task.path}:")
            tabbed_stderr = "\n".join(
                "\t" + line for line in result.stderr.splitlines()
            )
            logging.info(tabbed_stderr)
    logging.info("\n")

    all_passed = all(result.success for result in results)
    if logging.getLogger().isEnabledFor(logging.INFO) or detailed:
        print_test_output(results)
    if all_passed:
        logging.info("\x1b[32;1mALL TEST PASSED\x1b[0m")
    else:
        logging.warning("\x1b[31;1mSOME TESTS FAILED\x1b[0m")

    return all_passed, results


@contextlib.contextmanager
def cwd(path: Path) -> Generator[None, None, None]:
    old_wd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_wd)


def run_test_task(task: TestTask) -> TestResult:
    assignment_path = task.path.absolute()
    logging.debug(f"Running tests for {assignment_path}")
    if not is_valid_assignment(assignment_path):
        raise ValueError(f"{assignment_path} is not a valid TMC assignment")
    with cwd(assignment_path):
        logging.debug(f"{assignment_path=}, {os.getcwd()=}")
        result = subprocess.run(
            ["python3", "-m", "tmc"],
            cwd=assignment_path,
            capture_output=True,
            text=True,
        )
    logging.debug(f"Test run complete; {result.returncode=}")
    return TestResult(task, result.returncode == 0, result.stdout, result.stderr)


def main(argv: Optional[list[str]] = None) -> int:
    parser = argparse.ArgumentParser(
        "tmc-course",
        description="Helper for building TestMyCode python programming courses",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=importlib.metadata.version(__package__ or __name__),
    )

    verbosity_grp = parser.add_mutually_exclusive_group()
    verbosity_grp.add_argument(
        "--quiet", "-q", action="store_true", help="Only output warning"
    )
    verbosity_grp.add_argument(
        "--debug", "-d", action="store_true", help="Output debugging information"
    )

    actions = parser.add_subparsers(
        dest="action",
        required=True,
        metavar="ACTION",
    )

    # INIT
    init_grp = actions.add_parser(
        "init", help="Initialize a new course, part or assignment"
    )
    init_actions = init_grp.add_subparsers(
        dest="init_action", required=True, metavar="TYPE"
    )

    # INIT COURSE
    init_course_grp = init_actions.add_parser("course", help="Initialize a new course")
    init_course_grp.add_argument(
        "path", type=str, nargs="+", help="Path(s) of the new course(s)"
    )

    # INIT PART
    init_part_grp = init_actions.add_parser("part", help="Initialize a new course part")
    init_part_grp.add_argument(
        "path", type=str, nargs="+", help="Path(s) of the new part(s)"
    )

    # INIT ASSIGNMENT
    init_assignment_grp = init_actions.add_parser(
        "assignment", help="Initialize a assignment"
    )
    init_assignment_grp.add_argument(
        "path", type=str, nargs="+", help="Path(s) of the new assignment(s)"
    )
    language_grp = init_assignment_grp.add_mutually_exclusive_group(required=True)
    language_grp.add_argument(
        "-e", "--english", action="store_true", help="Use English language templates"
    )
    language_grp.add_argument(
        "-f", "--finnish", action="store_true", help="Use Finnish language templates"
    )

    # TEST
    test_grp = actions.add_parser("test", help="Test a new course, part or assignment")
    test_grp.add_argument(
        "path", type=str, nargs="+", help="Path(s) to test (course, part or assignment)"
    )
    test_grp.add_argument(
        "--details", action="store_true", help="Show more details about test results"
    )

    # UPDATE
    update_grp = actions.add_parser(
        "update", help="Update TMC-python-runner embedded in assignments"
    )
    update_grp.add_argument("path", type=str, help="Course root directory")

    # Verbosity control
    args = parser.parse_args(argv)
    if not (args.quiet or args.debug):
        logging.basicConfig(format="%(message)s", level=logging.INFO)
    if args.quiet:
        logging.basicConfig(format="%(message)s", level=logging.WARNING)
    if args.debug:
        logging.basicConfig(
            format="%(levelname)s:%(asctime)s: %(message)s", level=logging.DEBUG
        )

    try:
        if args.action == "init":
            paths = (Path(path) for path in args.path)
            if args.init_action == "course":
                for path in paths:
                    init_course(path)
            elif args.init_action == "part":
                for path in paths:
                    init_part(path.parent, path.name)
            else:  # args.init_action == "assignment"
                language: Literal["fi", "en"] = "fi" if args.finnish else "en"
                for path in paths:
                    init_assignment(
                        path.parent.parent, path.parent.name, path.name, language
                    )
        if args.action == "test":
            all_passed, _ = test(
                [Path(path) for path in args.path], detailed=args.details
            )
            if not all_passed:
                return 1
        if args.action == "update":
            update_course(Path(args.path))
    except ActionCancelledException:
        print("OK, quitting")
        return 1

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



import hmac
import hashlib

def write_hmac(secret, output):
    h = hmac.new(bytearray(secret, encoding='utf8'), output.encode(), hashlib.sha256)
    with open(".tmc_test_results.hmac.sha256", "w") as text_file:
      text_file.write(h.hexdigest())



import logging
import shutil
from pathlib import Path
from unittest.mock import ANY, call, patch

import pytest
import responses

from testing.util import assert_dir_equals, normalized_filecmp
from tmc_course import tmc_course


@pytest.mark.parametrize(
    "user_input, quit_default, continue_on_y",
    (
        ("", True, False),
        ("", True, True),
        ("y", True, False),
        ("y", False, False),
        ("n", True, True),
        ("n", False, True),
        ("Y", True, False),
        ("Y", False, False),
    ),
)
def test_check_from_user_quits(user_input, quit_default, continue_on_y, monkeypatch):
    monkeypatch.setattr("builtins.input", lambda _: user_input)
    with pytest.raises(tmc_course.ActionCancelledException):
        tmc_course.check_from_user("msg", quit_default, continue_on_y)


@pytest.mark.parametrize(
    "user_input, quit_default, continue_on_y",
    (
        ("", False, False),
        ("", False, True),
        ("y", True, True),
        ("y", False, True),
        ("n", True, False),
        ("n", False, False),
        ("Y", True, True),
        ("Y", False, True),
        ("N", True, False),
        ("N", False, False),
    ),
)
def test_check_from_user_continues(
    user_input, quit_default, continue_on_y, monkeypatch
):
    monkeypatch.setattr("builtins.input", lambda _: user_input)
    tmc_course.check_from_user("msg", quit_default, continue_on_y)


@pytest.mark.parametrize("resource_name", list(tmc_course.SkeletonFile))
def test_all_resources_exist(resource_name, tmp_path):
    tmp_file = tmp_path / "output.tmp"
    tmc_course.add_skeleton_file(resource_name, tmp_file)
    assert len(tmp_file.read_text()) > 1


def test_add_skeleton_file(tmp_path):
    tmp_file = tmp_path / "output.tmp"
    tmc_course.add_skeleton_file(
        tmc_course.SkeletonFile.COURSE_TMCPROJECT_YML, tmp_file
    )
    assert (
        tmp_file.read_text()
        == 'sandbox_image: "eu.gcr.io/moocfi-public/tmc-sandbox-python"\n'
    )


def test_init_course(tmp_path):
    course_path = tmp_path / "NewCourse"
    tmc_course.init_course(course_path)
    assert (course_path / ".gitignore").exists()
    assert (course_path / ".tmcproject.yml").exists()


def test_init_course_existing_checks_user(tmp_path, monkeypatch):
    course_path = tmp_path / "NewCourse"
    course_path.mkdir()
    with patch.object(tmc_course, "check_from_user") as mock:
        tmc_course.init_course(course_path)
        mock.assert_called_once()


@pytest.mark.parametrize("name", ("-foo-" "foo bar", "baz-bar", ".foo"))
def test_init_course_invalid_name(tmp_path, name):
    course_dir = tmp_path / name
    with pytest.raises(ValueError):
        tmc_course.init_course(course_dir)


@pytest.mark.parametrize("name", ("bar", "foo_bar", "foo1", "1_foo"))
def test_init_course_valid_name(tmp_path, name):
    tmc_course.init_course(tmp_path / name)


def test_is_valid_course(test_resource_path):
    assert tmc_course.is_valid_course(test_resource_path / "valid_course")


def test_is_valid_course_nonexistant(tmp_path):
    assert not tmc_course.is_valid_course(tmp_path / "no_such")


def test_is_valid_course_not_dir(tmp_path):
    filepath = tmp_path / "file.txt"
    filepath.touch()
    assert not tmc_course.is_valid_course(filepath)


def test_is_valid_course_no_tmcproject_yml(tmp_path):
    dirpath = tmp_path / "dir"
    dirpath.mkdir()
    assert not tmc_course.is_valid_course(dirpath)


def test_init_part(tmp_course):
    tmc_course.init_part(tmp_course, "part01")


def test_init_part_validates_course(tmp_path):
    course_path = tmp_path / "NewCourse"
    course_path.mkdir()
    with pytest.raises(ValueError):
        tmc_course.init_part(course_path, "part01")


@pytest.mark.parametrize("name", ("foo-bar", ".foo", "foo/bar"))
def test_init_part_checks_name(tmp_course, name):
    with pytest.raises(ValueError):
        tmc_course.init_part(tmp_course, name)


def test_init_part_checks_if_exists(tmp_course):
    (tmp_course / "part01").mkdir()
    with patch.object(tmc_course, "check_from_user") as mock:
        tmc_course.init_part(tmp_course, "part01")
        mock.assert_called_once()


def test_is_valid_part(test_resource_path):
    assert tmc_course.is_valid_part(test_resource_path / "valid_course" / "valid_part")


def test_is_valid_part_checks_exists(tmp_course):
    assert not tmc_course.is_valid_part(tmp_course / "no-such")


def test_is_valid_part_checks_is_dir(tmp_course):
    (tmp_course / "file.txt").touch()
    assert not tmc_course.is_valid_part(tmp_course / "file.txt")


def test_is_valid_part_in_course(tmp_path):
    (tmp_path / "course" / "part01").mkdir(parents=True, exist_ok=True)
    assert not tmc_course.is_valid_part(tmp_path / "course" / "part01")


def test_create_src_skeleton_fi(tmp_path, test_resource_path):
    tmc_course.create_src_skeleton(tmp_path, "fi")

    solution_file = tmp_path / "src" / "ratkaisu.py"
    expected = (
        test_resource_path
        / "valid_course"
        / "valid_part"
        / "valid_assignment_fi"
        / "src"
        / "ratkaisu.py"
    )
    assert solution_file.exists()
    assert solution_file.read_text() == expected.read_text()

    init_py = tmp_path / "src" / "__init__.py"
    assert init_py.exists()
    assert init_py.read_text() == ""


def test_create_src_skeleton_en(tmp_path, test_resource_path):
    tmc_course.create_src_skeleton(tmp_path, "en")

    solution_file = tmp_path / "src" / "solution.py"
    expected = (
        test_resource_path
        / "valid_course"
        / "valid_part"
        / "valid_assignment_en"
        / "src"
        / "solution.py"
    )
    assert solution_file.exists()
    assert solution_file.read_text() == expected.read_text()

    init_py = tmp_path / "src" / "__init__.py"
    assert init_py.exists()
    assert init_py.read_text() == ""


def test_create_src_skeleton_invalid_language(tmp_path):
    with pytest.raises(ValueError):
        tmc_course.create_src_skeleton(tmp_path, "nosuchlanguage")


def test_create_test_skeleton_fi(tmp_path, test_resource_path):
    tmc_course.create_test_skeleton(tmp_path, "valid_assignment_fi", "fi")

    test_file = tmp_path / "test" / "test_ratkaisu.py"
    expected = (
        test_resource_path
        / "valid_course"
        / "valid_part"
        / "valid_assignment_fi"
        / "test"
        / "test_ratkaisu.py"
    )
    assert test_file.exists()
    assert test_file.read_text() == expected.read_text()

    init_py = tmp_path / "test" / "__init__.py"
    assert init_py.exists()
    assert init_py.read_text() == ""


def test_create_test_skeleton_en(tmp_path, test_resource_path):
    tmc_course.create_test_skeleton(tmp_path, "valid_assignment_en", "en")

    test_file = tmp_path / "test" / "test_solution.py"
    expected = (
        test_resource_path
        / "valid_course"
        / "valid_part"
        / "valid_assignment_en"
        / "test"
        / "test_solution.py"
    )
    assert test_file.exists()
    assert test_file.read_text() == expected.read_text()

    init_py = tmp_path / "test" / "__init__.py"
    assert init_py.exists()
    assert init_py.read_text() == ""


def test_create_test_skeleton_invalid_language(tmp_path):
    with pytest.raises(ValueError):
        tmc_course.create_test_skeleton(tmp_path, "solution", "nosuchlanguage")


@responses.activate
def test_download_tmc_python_tester(test_resource_path, tmp_path):
    url = (
        "https://github.com/testmycode/tmc-python-tester/archive/refs/heads/master.zip"
    )
    zip_resource = test_resource_path / "tmc-python-tester.zip"
    with zip_resource.open("rb") as zip_handle:
        responses.get(url=url, body=zip_handle.read())
        tmc_course.download_tmc_python_tester(tmp_path, update=True)
    responses.assert_call_count(url, 1)
    assert normalized_filecmp(zip_resource, tmp_path / "tmc-python-tester.zip")


def test_download_tmc_python_tester_skips_no_update(tmp_path):
    with patch.object(tmc_course, "requests") as mock:
        (tmp_path / "tmc-python-tester.zip").touch()
        tmc_course.download_tmc_python_tester(tmp_path, update=False)
        mock.assert_not_called()


@responses.activate
def test_download_tmc_python_tester_updates(test_resource_path, tmp_path):
    (tmp_path / "tmc-python-tester.zip").touch()

    url = (
        "https://github.com/testmycode/tmc-python-tester/archive/refs/heads/master.zip"
    )
    zip_resource = test_resource_path / "tmc-python-tester.zip"
    with zip_resource.open("rb") as zip_handle:
        responses.get(url=url, body=zip_handle.read())
        tmc_course.download_tmc_python_tester(tmp_path, update=True)

    responses.assert_call_count(url, 1)
    assert normalized_filecmp(zip_resource, tmp_path / "tmc-python-tester.zip")


@responses.activate
def test_create_tmc_dir(test_resource_path, tmp_part):
    expected_path = (
        test_resource_path
        / "valid_course"
        / "valid_part"
        / "valid_assignment_en"
        / "tmc"
    )
    assg_path = tmp_part / "assg01"
    assg_path.mkdir()

    url = (
        "https://github.com/testmycode/tmc-python-tester/archive/refs/heads/master.zip"
    )
    zip_resource = test_resource_path / "tmc-python-tester.zip"
    with zip_resource.open("rb") as zip_handle:
        responses.get(url=url, body=zip_handle.read())
        tmc_course.create_tmc_dir(assg_path)

    assert_dir_equals(expected_path, assg_path / "tmc")


@responses.activate
def test_init_assignment_en(test_resource_path, tmp_part):
    url = (
        "https://github.com/testmycode/tmc-python-tester/archive/refs/heads/master.zip"
    )
    zip_resource = test_resource_path / "tmc-python-tester.zip"
    with zip_resource.open("rb") as zip_handle:
        responses.get(url=url, body=zip_handle.read())
        tmc_course.init_assignment(
            tmp_part.parent, tmp_part.name, "valid_assignment_en", "en"
        )

    assert_dir_equals(
        test_resource_path / "valid_course" / "valid_part" / "valid_assignment_en",
        tmp_part / "valid_assignment_en",
        ignore=["__pycache__"],
    )


def test_init_assignment_invalid_part(tmp_course):
    (tmp_course / "part1.txt").touch()
    with pytest.raises(ValueError):
        tmc_course.init_assignment(tmp_course, "part1.txt", "assg1", "fi")


def test_init_assignment_invalid_course(tmp_path):
    (tmp_path / "course" / "part1").mkdir(parents=True, exist_ok=True)
    with pytest.raises(ValueError):
        tmc_course.init_assignment(tmp_path / "course", "part1", "assg1", "fi")


@responses.activate
def test_init_assignment_fi(test_resource_path, tmp_part):
    url = (
        "https://github.com/testmycode/tmc-python-tester/archive/refs/heads/master.zip"
    )
    zip_resource = test_resource_path / "tmc-python-tester.zip"
    with zip_resource.open("rb") as zip_handle:
        responses.get(url=url, body=zip_handle.read())
        tmc_course.init_assignment(
            tmp_part.parent, tmp_part.name, "valid_assignment_fi", "fi"
        )

    assert_dir_equals(
        test_resource_path / "valid_course" / "valid_part" / "valid_assignment_fi",
        tmp_part / "valid_assignment_fi",
        ignore=["__pycache__"],
    )


@pytest.mark.parametrize("name", (".foo", "foo bar", "foo/bar", "foo-bar"))
def test_init_assignment_invalid_name(tmp_part, name):
    with pytest.raises(ValueError):
        tmc_course.init_assignment(tmp_part.parent, tmp_part.name, name, "en")


def test_init_assignment_tgt_exists_asks_user(tmp_part):
    (tmp_part / "assg1").mkdir()
    with patch.object(tmc_course, "check_from_user") as mock:
        tmc_course.init_assignment(tmp_part.parent, tmp_part.name, "assg1", "en")
        mock.assert_called_once()


def test_is_valid_assignment(test_resource_path):
    assert tmc_course.is_valid_assignment(
        test_resource_path / "valid_course" / "valid_part" / "valid_assignment_en"
    )


def test_is_valid_assignment_not_exist(tmp_path):
    assert not tmc_course.is_valid_assignment(tmp_path / "nosuch")


def test_is_valid_assignment_not_dir(tmp_path):
    (tmp_path / "file.txt").touch()
    assert not tmc_course.is_valid_assignment(tmp_path / "file.txt")


def test_is_valid_assignment_no_tmcprojectyml(tmp_path):
    (tmp_path / "assg").mkdir()
    assert not tmc_course.is_valid_assignment(tmp_path / "assg")


@responses.activate
def test_update_course(tmp_course, test_resource_path):
    url = (
        "https://github.com/testmycode/tmc-python-tester/archive/refs/heads/master.zip"
    )
    zip_resource = test_resource_path / "tmc-python-tester.zip"
    with zip_resource.open("rb") as zip_handle:
        responses.get(url=url, body=zip_handle.read())

        tmc_course.init_part(tmp_course, "part01")
        tmc_course.init_assignment(tmp_course, "part01", "valid_assignment_fi", "fi")
        tmc_course.init_assignment(tmp_course, "part01", "valid_assignment_en", "en")
        tmc_course.init_assignment(tmp_course, "part01", "assg03", "en")

        tmc_course.init_part(tmp_course, "part02")
        tmc_course.init_assignment(tmp_course, "part02", "valid_assignment_fi", "fi")
        tmc_course.init_assignment(tmp_course, "part02", "valid_assignment_en", "en")

    # add non-assignment dir
    (tmp_course / "part01" / "utils" / "src").mkdir(parents=True)
    (tmp_course / "part01" / "utils" / "src" / "file.txt").touch()

    # modify /tmc/
    (tmp_course / "part01" / "valid_assignment_fi" / "tmc" / "hmac_writer.py").unlink()
    (tmp_course / "part01" / "valid_assignment_en" / "tmc" / "runner.py").unlink()
    (tmp_course / "part01" / "assg03" / "tmc" / "hmac_writer.py").unlink()
    (tmp_course / "part02" / "valid_assignment_fi" / "tmc" / "reflect.py").unlink()
    (tmp_course / "part02" / "valid_assignment_en" / "tmc" / "utils.py").unlink()

    (tmp_course / "part01" / "valid_assignment_en" / "tmc" / "points.py").open(
        "w"
    ).write("MODIFIED")
    (tmp_course / "part02" / "valid_assignment_fi" / "tmc" / "points.py").open(
        "w"
    ).write("MODIFIED")

    # modify /src/ and /test/
    (tmp_course / "part01" / "assg03" / "src" / "solution.py").open("w").write(
        "DO NOT OVERWRITE"
    )
    (tmp_course / "part01" / "assg03" / "test" / "test_solution.py").open("w").write(
        "TEST DATA"
    )

    with zip_resource.open("rb") as zip_handle:
        responses.get(url=url, body=zip_handle.read())
        tmc_course.update_course(tmp_course)

    # Updates all in a part
    assert_dir_equals(
        test_resource_path
        / "valid_course"
        / "valid_part"
        / "valid_assignment_fi"
        / "tmc",
        tmp_course / "part01" / "valid_assignment_fi" / "tmc",
        ignore=["__pycache__"],
    )
    assert_dir_equals(
        test_resource_path
        / "valid_course"
        / "valid_part"
        / "valid_assignment_en"
        / "tmc",
        tmp_course / "part01" / "valid_assignment_en" / "tmc",
        ignore=["__pycache__"],
    )
    assert_dir_equals(
        test_resource_path
        / "valid_course"
        / "valid_part"
        / "valid_assignment_en"
        / "tmc",
        tmp_course / "part01" / "assg03" / "tmc",
        ignore=["__pycache__"],
    )

    # Updates all parts
    assert_dir_equals(
        test_resource_path
        / "valid_course"
        / "valid_part"
        / "valid_assignment_en"
        / "tmc",
        tmp_course / "part02" / "valid_assignment_en" / "tmc",
        ignore=["__pycache__"],
    )

    # Doesn't touch /src/ or /test/
    assert (
        tmp_course / "part01" / "assg03" / "src" / "solution.py"
    ).read_text() == "DO NOT OVERWRITE"
    assert (
        tmp_course / "part01" / "assg03" / "test" / "test_solution.py"
    ).read_text() == "TEST DATA"

    # Didn't touch the util dir
    assert len(list((tmp_course / "part01" / "utils" / "src").iterdir())) == 1
    assert len(list((tmp_course / "part01" / "utils").iterdir())) == 1


def test_main_init_course(tmp_path):
    course_paths = [tmp_path / "NewCourse1", tmp_path / "NewCourse2"]
    with patch.object(tmc_course, "init_course") as mock:
        tmc_course.main(["init", "course", str(course_paths[0]), str(course_paths[1])])
        mock.assert_has_calls(
            [
                call(course_paths[0]),
                call(course_paths[1]),
            ],
            any_order=True,
        )


def test_main_init_part(tmp_course):
    part_paths = [tmp_course / "part01", tmp_course / "part02"]
    with patch.object(tmc_course, "init_part") as mock:
        tmc_course.main(["init", "part", str(part_paths[0]), str(part_paths[1])])
        mock.assert_has_calls(
            [
                call(tmp_course, "part01"),
                call(tmp_course, "part02"),
            ],
            any_order=True,
        )


def test_main_init_assignment(tmp_course):
    assg_paths = [tmp_course / "part01" / "assg1", tmp_course / "part01" / "assg2"]
    with patch.object(tmc_course, "init_assignment") as mock:
        tmc_course.main(
            ["init", "assignment", str(assg_paths[0]), str(assg_paths[1]), "-f"]
        )
        mock.assert_has_calls(
            [
                call(tmp_course, "part01", "assg1", "fi"),
                call(tmp_course, "part01", "assg2", "fi"),
            ],
            any_order=True,
        )


def test_main_init_assignment_en(tmp_course):
    assg_paths = [tmp_course / "part01" / "assg1", tmp_course / "part01" / "assg2"]
    with patch.object(tmc_course, "init_assignment") as mock:
        tmc_course.main(
            ["init", "assignment", str(assg_paths[0]), str(assg_paths[1]), "-e"]
        )
        mock.assert_has_calls(
            [
                call(tmp_course, "part01", "assg1", "en"),
                call(tmp_course, "part01", "assg2", "en"),
            ],
            any_order=True,
        )


def test_main_update(tmp_course):
    with patch.object(tmc_course, "update_course") as mock:
        tmc_course.main(["update", str(tmp_course)])
        mock.assert_called_once_with(tmp_course)


def test_verbosity_quiet():
    with patch.object(tmc_course, "update_course"):
        with patch.object(tmc_course.logging, "basicConfig") as mock_logging:
            tmc_course.main(["--quiet", "update", "nosuchcourse"])
            mock_logging.assert_called_once_with(format=ANY, level=logging.WARNING)


def test_verbosity_normal():
    with patch.object(tmc_course, "update_course"):
        with patch.object(tmc_course.logging, "basicConfig") as mock_logging:
            tmc_course.main(["update", "nosuchcourse"])
            mock_logging.assert_called_once_with(format=ANY, level=logging.INFO)


def test_verbosity_debug():
    with patch.object(tmc_course, "update_course"):
        with patch.object(tmc_course.logging, "basicConfig") as mock_logging:
            tmc_course.main(["--debug", "update", "nosuchcourse"])
            mock_logging.assert_called_once_with(format=ANY, level=logging.DEBUG)


def test_test_task_properties():
    task = tmc_course.TestTask(Path("Course/Part/Assignment"))
    assert task.course_path == Path("Course")
    assert task.part_path == Path("Course/Part")


def test_collect_tasks(test_resource_path):
    tasks = list(
        tmc_course.collect_tasks(
            [
                test_resource_path / "test_runner_test_all_pass",
                test_resource_path / "test_runner_test_some_pass",
                test_resource_path / "test_runner_test_all_fail" / "part01",
            ]
        )
    )
    assert len(tasks) == 10
    assert set(task.path for task in tasks) == {
        test_resource_path / "test_runner_test_all_pass" / "part01" / "assg01",
        test_resource_path / "test_runner_test_all_pass" / "part01" / "assg02",
        test_resource_path / "test_runner_test_all_pass" / "part02" / "assg03",
        test_resource_path / "test_runner_test_all_pass" / "part02" / "assg04",
        test_resource_path / "test_runner_test_some_pass" / "part01" / "assg01",
        test_resource_path / "test_runner_test_some_pass" / "part01" / "assg02",
        test_resource_path / "test_runner_test_some_pass" / "part02" / "assg03",
        test_resource_path / "test_runner_test_some_pass" / "part02" / "assg04",
        test_resource_path / "test_runner_test_all_fail" / "part01" / "assg01",
        test_resource_path / "test_runner_test_all_fail" / "part01" / "assg02",
    }


def test_collect_tasks_invalid(tmp_path):
    tasks = list(tmc_course.collect_tasks([tmp_path]))
    assert len(tasks) == 0


def test_run_test_task_invalid_path(tmp_path):
    with pytest.raises(ValueError):
        tmc_course.run_test_task(tmc_course.TestTask(tmp_path / "no_such"))


def test_test_course(test_resource_path, tmp_path):
    shutil.copytree(
        test_resource_path / "test_runner_test_all_pass",
        tmp_path / "test_runner_test_all_pass",
    )
    success, results = tmc_course.test([tmp_path / "test_runner_test_all_pass"])
    assert success
    assert len(results) == 4


def test_test_part(test_resource_path, tmp_path):
    shutil.copytree(
        test_resource_path / "test_runner_test_all_pass",
        tmp_path / "test_runner_test_all_pass",
    )
    success, results = tmc_course.test(
        [tmp_path / "test_runner_test_all_pass" / "part01"]
    )
    assert success
    assert len(results) == 2


def test_test_assg(test_resource_path, tmp_path):
    shutil.copytree(
        test_resource_path / "test_runner_test_all_pass",
        tmp_path / "test_runner_test_all_pass",
    )
    success, results = tmc_course.test(
        [tmp_path / "test_runner_test_all_pass" / "part01" / "assg01"]
    )
    assert success
    assert len(results) == 1


def test_test_multiple_assgs(test_resource_path, tmp_path):
    shutil.copytree(
        test_resource_path / "test_runner_test_all_pass",
        tmp_path / "test_runner_test_all_pass",
    )
    success, results = tmc_course.test(
        [
            tmp_path / "test_runner_test_all_pass" / "part01" / "assg01",
            tmp_path / "test_runner_test_all_pass" / "part01" / "assg02",
        ]
    )
    assert success
    assert len(results) == 2


def test_test_multiple_parts(test_resource_path, tmp_path):
    shutil.copytree(
        test_resource_path / "test_runner_test_all_pass",
        tmp_path / "test_runner_test_all_pass",
    )
    success, results = tmc_course.test(
        [
            tmp_path / "test_runner_test_all_pass" / "part01",
            tmp_path / "test_runner_test_all_pass" / "part02",
        ]
    )
    assert success
    assert len(results) == 4


def test_test_multiple_mixture(test_resource_path, tmp_path):
    shutil.copytree(
        test_resource_path / "test_runner_test_all_pass",
        tmp_path / "test_runner_test_all_pass",
    )
    success, results = tmc_course.test(
        [
            tmp_path / "test_runner_test_all_pass" / "part01",
            tmp_path / "test_runner_test_all_pass" / "part02" / "assg03",
        ]
    )
    assert success
    assert len(results) == 3


def test_test_all_fail(test_resource_path, tmp_path):
    shutil.copytree(
        test_resource_path / "test_runner_test_all_fail",
        tmp_path / "test_runner_test_all_fail",
    )
    success, results = tmc_course.test(
        [test_resource_path / "test_runner_test_all_fail"]
    )
    assert not success
    assert len(results) == 4


def test_test_some_pass(test_resource_path, tmp_path):
    shutil.copytree(
        test_resource_path / "test_runner_test_some_pass",
        tmp_path / "test_runner_test_some_pass",
    )
    success, results = tmc_course.test([tmp_path / "test_runner_test_some_pass"])
    assert not success
    assert len(results) == 4


def test_test_multi(test_resource_path, tmp_path):
    shutil.copytree(
        test_resource_path / "test_runner_test_some_pass",
        tmp_path / "test_runner_test_some_pass",
    )
    shutil.copytree(
        test_resource_path / "test_runner_test_all_pass",
        tmp_path / "test_runner_test_all_pass",
    )
    success, results = tmc_course.test(
        [
            tmp_path / "test_runner_test_some_pass",
            tmp_path / "test_runner_test_all_pass",
        ]
    )
    assert not success
    assert len(results) == 8


def test_main_test(tmp_course):
    paths = [tmp_course / "part01", tmp_course / "part02"]
    with patch.object(tmc_course, "test") as mock:
        mock.return_value = (True, [])
        res = tmc_course.main(["test", str(paths[0]), str(paths[1])])
        mock.assert_called_once_with(
            [
                paths[0],
                paths[1],
            ],
            detailed=False,
        )
        assert res == 0


def test_main_test_detailed(tmp_course):
    paths = [tmp_course / "part01", tmp_course / "part02"]
    with patch.object(tmc_course, "test") as mock:
        mock.return_value = (False, [])
        res = tmc_course.main(["test", str(paths[0]), str(paths[1]), "--details"])
        mock.assert_called_once_with(
            [
                paths[0],
                paths[1],
            ],
            detailed=True,
        )
        assert res == 1



from unittest import TextTestRunner, TestLoader
from .result import TMCResult
from .points import _parse_points, _name_test
from itertools import chain
import json


class TMCTestRunner(TextTestRunner):
    """A test runner for TMC exercises.
    """

    resultclass = TMCResult

    def __init__(self, *args, **kwargs):
        super(TMCTestRunner, self).__init__(*args, **kwargs)

    def run(self, test):
        print('Running tests with some TMC magic...')
        return super(TMCTestRunner, self).run(test)

    def available_points(self):
        testLoader = TestLoader()
        tests = testLoader.discover('.', 'test*.py', None)
        try:
            tests = list(chain(*chain(*tests._tests)))
        except Exception as error:
            print("Received following Exception:", error)
            tests.debug()

        points = map(_parse_points, tests)
        names = map(_name_test, tests)

        result = dict(zip(names, points))

        with open('.available_points.json', 'w') as f:
            json.dump(result, f, ensure_ascii=False)
